#!/usr/bin/env python3
# This file is part of GUFI, which is part of MarFS, which is released
# under the BSD license.
#
#
# Copyright (c) 2017, Los Alamos National Security (LANS), LLC
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation and/or
# other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#
# From Los Alamos National Security, LLC:
# LA-CC-15-039
#
# Copyright (c) 2017, Los Alamos National Security, LLC All rights reserved.
# Copyright 2017. Los Alamos National Security, LLC. This software was produced
# under U.S. Government contract DE-AC52-06NA25396 for Los Alamos National
# Laboratory (LANL), which is operated by Los Alamos National Security, LLC for
# the U.S. Department of Energy. The U.S. Government has rights to use,
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR LOS
# ALAMOS NATIONAL SECURITY, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is
# modified to produce derivative works, such modified software should be
# clearly marked, so as not to confuse it with the version available from
# LANL.

# THIS SOFTWARE IS PROVIDED BY LOS ALAMOS NATIONAL SECURITY, LLC AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL LOS ALAMOS NATIONAL SECURITY, LLC OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
# OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
# OF SUCH DAMAGE.



import argparse
import os
import subprocess
import sys

import bee_common
import bee_config

# hiding traceback
sys.tracebacklimit=0

#        1024  1000  1024
SIZES = ['K',  'KB', 'KiB',
         'M',  'MB', 'MiB',
         'G',  'GB', 'GiB',
         'T',  'TB', 'TiB',
         'P',  'PB', 'PiB',
         'E',  'EB', 'EiB',
         'Z',  'ZB', 'ZiB',
         'Y',  'YB', 'YiB']

def select_size(args, name):
    # in GNU ls, order of arguments determines which one has precedence
    # here, --human-readable has precedence
    if args.human_readable:
        column = 'human_readable_size({}, ' + str(args.size_width) + ')'
    elif args.blocksize:
        column = 'blocksize({}, "' + args.blocksize + '", ' + str(args.size_width) + ')'
    else:
        column = 'printf("%' + str(args.size_width) + 'd", {})'

    return [column.format(name)]

TIME_STYLES = {
    'full-iso' : '%F %T %z',
    'long-iso' : '%F %T',
    'iso'      : '%m-%d %R',
    'locale'   : '%b  %e %H:%M',
}

def select_time(style, name):
    if not style:
        style = 'locale'

    try:
        column = 'strftime("' + TIME_STYLES[style] + '", {})'
    except:
        raise ValueError('bee ls: invalid argument \'{}\' for \'time style\'\nExpected time styles: \'{}\''.format(style, TIME_STYLES))

    return [column.format(name)]

def build_select(args):
    '''Build the SELECT portion of the query on the initial scan of entries'''
    select = []

    # ordering of concatenation matters

    if args.inode:
        select += ['inode']

    if args.size:
        select += ['blocks']

    if args.long_listing:
        select += ['modetotxt(mode)', 'printf("%' + str(args.nlink_width) + 'd", nlink)', 'uidtouser(uid, ' + str(args.user_width) + ')']
        if not args.no_group:
            select += ['gidtogroup(gid, ' + str(args.group_width) + ')']
        select += select_size(args, 'size') + select_time(args.time_style, 'mtime') + ['case when (type == "l") then (name || " -> " || linkname) else name end']

    else:
        # at minimum, print the file name
        select += ['name']

    return select

def build_recursive_select(args):
    '''Build the SELECT portion of the query on the initial scan of entries'''
    select = []

    # ordering of concatenation matters

    if args.inode:
        select += ['uinode']

    if args.size:
        select += ['ublocks']

    if args.long_listing:
        select += ['modetotxt(umode)', 'printf("%' + str(args.nlink_width) + 'd", unlink)', 'uidtouser(uuid, ' + str(args.user_width) + ')']
        if not args.no_group:
            select += ['gidtogroup(ugid, ' + str(args.group_width) + ')']
        select += select_size(args, 'usize') + select_time(args.time_style, 'umtime') + ['case when (utype == "l") then (ufullpath || "/" || uname || " -> " || ulinkname) else ufullpath || "/" || uname end']

    else:
        # at minimum, print the file name
        select += ['ufullpath || "/" || uname']

    return select

def build_where(args, name = None):
    '''Build the WHERE portion of the query'''
    where = []

    if name:
        where = ['name REGEXP "^{}$"'.format(name)]
    else:
        # -a: if show all, override -A
        if args.all:
            pass
        else:
            # -A: ignore . and ..
            if args.almost_all:
                where += ['name <> "." AND name <> ".."']

            # everything that's not hidden
            else:
                where += ['name REGEXP r"^(?![.]).*$"']

        # -B: ignore files that end with ~
        if args.ignore_backups == True:
            where += ['name REGEXP r"^.*(?<![~])$"']

    return where

def build_order_by(args):
    '''Build the ORDER BY portion of the query.'''
    order_by = []

    if args.sort_largest:
        order_by += ['size ' + ('ASC' if args.reverse else 'DESC')]

    if args.mtime:
        order_by += ['mtime ' + ('ASC' if args.reverse else 'DESC')]

    if len(order_by) == 0:
        order_by += ['name COLLATE NOCASE ' + ('DESC' if args.reverse else 'ASC')]

    if args.no_sort:
        order_by = []

    return order_by

def build_order_by_recursive(args):
    '''Build the ORDER BY portion of the query.'''
    order_by = []

    if args.sort_largest:
        order_by += ['usize ' + ('ASC' if args.reverse else 'DESC')]

    if args.mtime:
        order_by += ['umtime ' + ('ASC' if args.reverse else 'DESC')]

    if len(order_by) == 0:
        order_by += ['ufullpath || "/" || uname COLLATE NOCASE ' + ('DESC' if args.reverse else 'ASC')]

    if args.no_sort:
        order_by = []

    return order_by

# argv[0] should be the command name
def exec_index(argv, config_path):
    # storing name of test config file
    test_config_path ="config.test"

    # find and parse the configuration file first
    if(config_path == bee_config.DEFAULT_PATH or config_path == test_config_path):
        config = bee_config.Server(config_path)
    else:
        config = bee_config.envServer(bee_config.DEFAULT_PATH)

    class FullTimeAction(argparse.Action):
        def __init__(self, option_strings, dest, nargs=None, **kwargs):
            super(FullTimeAction, self).__init__(option_strings, dest, 0, **kwargs)
        def __call__(self, parser, namespace, values, option_string=None):
            # --full-time
            #        like -l --time-style=full-iso
            setattr(namespace, 'long_listing', True)
            setattr(namespace, 'time_style', 'full-iso')

    # parse the arguments
    parser = argparse.ArgumentParser(description='BeeGFS Hive index version of create', add_help=False)
    # override help to not use -h
    parser.add_argument('--help', '-h',                                           action='help',                                      help='show this help message and exit')
    parser.add_argument('-X',                   dest='max_memory_usage',    action='store',     type=str,                       help='Max memory usage e.g 8M, 1G, 8GB, 16gb', default='1G')
    parser.add_argument('-F',                   dest='fs_path',             action='store',     type=str,                       help='File system path for which index will be created', required=False)
    parser.add_argument('-I',                   dest='index_path',          action='store',     type=str,                       help='Index directory path', default='')
    parser.add_argument('-n',                   dest='num_threads',         action='store',     type=int,                       help='Number of threads to create index', default=1)
    parser.add_argument('-s',                   dest='create_ts',           action='store_true',                                help='Create tree summary table along with other tables')
    parser.add_argument('-x',                   dest='xattars',             action='store_true',                                help='Pull xattars from source')
    parser.add_argument('-z',                   dest='max_level',           action='store',     type=int,                       help='Max level to go down')
    parser.add_argument('-S',                   dest='create_only_ts',      action='store_true',                                help='Create only tree summary table')
    parser.add_argument('--version', '-v',                                  action='version',          version=os.path.basename(os.path.realpath(__file__)) + ' 2.0.0')
    parser.add_argument('-C',                   dest='print_scanned_dirs',  action='store_true',                                help='print the number of scanned directories')
    parser.add_argument('-M',                   dest='mnt_path',            action='store',     type=str,                       help='File system mount point path', default='')
    parser.add_argument('-p','--port',          dest='port_num',            action='store',     type=str,                       help='port number to connect with client', default='6000')
    parser.add_argument('-V','--verbose',       dest='debug_mode',          action='store',     type=str,                       help='enable/disable debugging by giving 1/0', default='0')
    parser.add_argument('-k',                   dest='create_index',        action='store_true',                                help='Run the create-index')
    parser.add_argument('-U',                   dest='update_index',        action='store_true',                                help='Run the update-index')
    parser.add_argument('-B',                   dest='no_beegfs_meta',        action='store_true',                                help='Do not extract BeeGFS specific metadata')

    args = parser.parse_args(argv[2:]);

    query = []
    paths = []
    query_cmd = []

    if(args.debug_mode == "1"):
        print("debugging enabled\n")
    else:
        print("debugging not enabled\n")

    # Get the fs_path and index_directory_path from config file.
    paths = ['']
    if(config_path == bee_config.DEFAULT_PATH or config_path == test_config_path):
        cpath = config.indexpaths()
        mntpath = config.mountpath()
    else:
        cpath = config_path

    fspath = [os.path.normpath(os.path.sep.join([cpath, path])) for path in paths]
    indexpaths = fspath[0].split(':')
    config_srcpath = indexpaths[0]
    if len(indexpaths) == 2:
        config_idxPath = indexpaths[1]
    else:
        config_idxPath = ""

    # Storing port number into variable, if -p,--port is not passed from cmdline, using default port number
    port = args.port_num

    # Checking length of port, the range of port number is from 0 to 65535
    if( len (args.port_num) > 5):
        if(args.debug_mode == "1"):
            print("Please provide correct port number\n")
        sys.exit()

    # Get mount point path from config file
    for path in paths:
        mnt_path = os.path.normpath(os.path.sep.join([mntpath, path]))

    # If -M i.e Filesystem mount point path is not passed from cmdline , use MountPath from config file.
    if args.mnt_path == "" or args.mnt_path == None:
        args.mnt_path = mnt_path

    # If -F i.e Filesystem path is not passed from cmdline , use FSpath from config file.
    if args.fs_path == "" or args.fs_path == None:
        args.fs_path = config_srcpath

    if args.fs_path == "":
       print("No FSPath passed from cmdline or available in the config file")
       exit(1)

    # If index is already created and now need to create only treesummary table
    if args.create_only_ts:
        if args.index_path == "" and config_idxPath == "":
            print(("Only Create Tree summary table for FSPath: ", args.fs_path))
            exec_path = "%s/bin/index/bee-create-ts-table" % config.executable()
            query_cmd = [exec_path, '-X', str(args.max_memory_usage), '-n', str(args.num_threads), args.fs_path]

        else:
            print(("Only create Tree summary table for index directory: ", args.index_path))
            # If -I is not passed but config file has index directory path, use that path for index directory.
            if args.index_path == "":
                args.index_path = config_idxPath
            exec_path = "%s/bin/index/bee-create-ts-table" % config.executable()
            query_cmd = [exec_path, '-X', str(args.max_memory_usage), '-n', str(args.num_threads), args.index_path]

    else:

        # If -I is not passed from cmdline and config file also do not have index directory path, create intree index.
        if args.index_path == "" and config_idxPath == "":
            index_path = args.fs_path
            if (args.debug_mode == "1"):
                print(("create and update the intree index: ", args.fs_path))
            exec_path = "%s/bin/index/bee-index" % config.executable()
            query_cmd = [exec_path, args.fs_path, args.fs_path, mnt_path, port]
        else:
            # If -I is not passed but config file has index directory path, use that path for index directory.
            if args.index_path == "":
                args.index_path = config_idxPath
                if (args.debug_mode == "1"):
                    print(("create and update the outtree index: ", args.index_path))
            exec_path = "%s/bin/index/bee-index" % config.executable()
            index_path = args.index_path
            query_cmd = [exec_path, args.fs_path, index_path, mnt_path, port]

        # If -V is passed from cli then append it with query_cmd
        if args.debug_mode:
            query_cmd += [str(args.debug_mode)]

        # If -X is passed from cli then append it with query_cmd
        if args.max_memory_usage:
            query_cmd += ['-X', str(args.max_memory_usage)]

        # If -n is passed from cli then append it with query_cmd
        if args.num_threads:
            query_cmd += ['-n', str(args.num_threads)]

        # If -z is passed from cli then append it with query_cmd
        if args.max_level:
            query_cmd += ['-z', str(args.max_level)]

        # If -s is passed from cli then append it with query_cmd
        if args.create_ts:
            query_cmd += ['-s']

        # If -x is passed from cli then append it with query_cmd
        if args.xattars:
            query_cmd += ['-x']

        # If -C is passed from cli then append it with query_cmd
        if args.print_scanned_dirs:
            query_cmd += ['-C']

        #If -k is passed from cli then append it with query_cmd
        if args.create_index:
            query_cmd += ['-k']

        #If -U is passed from cli then append it with query_cmd
        if args.update_index:
            query_cmd += ['-U']

        #If -B is passed from cli then do not extractBbeeGFS specific metadata
        if args.no_beegfs_meta:
            query_cmd += ['-B']

    query = subprocess.Popen(query_cmd)

    query.communicate() # block until query finishes

    return query.returncode
