#!/usr/bin/env python3
# This file is part of GUFI, which is part of MarFS, which is released
# under the BSD license.
#
#
# Copyright (c) 2017, Los Alamos National Security (LANS), LLC
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation and/or
# other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#
# From Los Alamos National Security, LLC:
# LA-CC-15-039
#
# Copyright (c) 2017, Los Alamos National Security, LLC All rights reserved.
# Copyright 2017. Los Alamos National Security, LLC. This software was produced
# under U.S. Government contract DE-AC52-06NA25396 for Los Alamos National
# Laboratory (LANL), which is operated by Los Alamos National Security, LLC for
# the U.S. Department of Energy. The U.S. Government has rights to use,
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR LOS
# ALAMOS NATIONAL SECURITY, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is
# modified to produce derivative works, such modified software should be
# clearly marked, so as not to confuse it with the version available from
# LANL.

# THIS SOFTWARE IS PROVIDED BY LOS ALAMOS NATIONAL SECURITY, LLC AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL LOS ALAMOS NATIONAL SECURITY, LLC OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
# OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
# OF SUCH DAMAGE.



import argparse
import os
import subprocess
import sys

import bee_common
import bee_config

# hiding traceback
sys.tracebacklimit=0

# argv[0] should be the command name
def exec_update(argv, config_path):
    # storing name of test config file
    test_config_path ="config.test"

    # find and parse the configuration file first
    if(config_path == bee_config.DEFAULT_PATH or config_path == test_config_path):
        config = bee_config.Server(config_path)
    else:
        config = bee_config.envServer(bee_config.DEFAULT_PATH)

    # parse the arguments
    parser = argparse.ArgumentParser(description='BeeGFS Hive index version of update', add_help=False)
    # override help to not use -h
    parser.add_argument('--help',                                           action='help',                                      help='show this help message and exit')
    parser.add_argument('-F',                   dest='fs_path',             action='store',     type=str,                       help='File system path for which index will be created', required=True)
    parser.add_argument('-I',                   dest='index_path',          action='store',     type=str,                       help='Index directory path', default='')
    parser.add_argument('-M',                   dest='mnt_path',            action='store',     type=str,                       help='File system mount point path', default='')
    parser.add_argument('-p','--port',          dest='port_num',            action='store',     type=str,                       help='port number to connect with client', default='6000')
    parser.add_argument('-V','--verbose',       dest='debug_mode',          action='store',     type=str,                       help='enable/disable debugging by giving 1/0')
    parser.add_argument('-B',                   dest='no_beegfs_meta',      action='store',     type=str,                       help='Do not extract BeeGFS metadata')

    args = parser.parse_args(argv[2:])

    query = []

    paths = []

    if(args.debug_mode == "1"):
        print("debugging enabled\n")
    else:
        print("debugging not enabled\n")

    # Get the fs_path and index_directory_path from config file.
    paths = ['']
    if(config_path == bee_config.DEFAULT_PATH or config_path == test_config_path):
        cpath = config.indexpaths()
        mntpath = config.mountpath()
    else:
        cpath = config_path

    fspath = [os.path.normpath(os.path.sep.join([cpath, path])) for path in paths]
    indexpaths = fspath[0].split(':')
    config_srcpath = indexpaths[0]

    if len(indexpaths) == 2:
        config_idxPath = indexpaths[1]
    else:
        config_idxPath = ""

    # Storing port number into variable, if -p,--port is not passed from cmdline, using default port number
    port = args.port_num

    # Checking length of port, the range of port number is from 0 to 65535
    if( len (args.port_num) > 5):
        if(args.debug_mode == "1"):
            print("Please provide correct port number\n")
        sys.exit()

    # Get mount point path from config file
    for path in paths:
        mnt_path = os.path.normpath(os.path.sep.join([mntpath, path]))

    # If -M i.e Filesystem mount point path is not passed from cmdline , use MountPath from config file.
    if args.mnt_path == "" or args.mnt_path == None:
        args.mnt_path = mnt_path

    # If -F i.e Filesystem path is not passed from cmdline , use FSpath from config file.
    if args.fs_path == "":
        args.fs_path = config_srcpath

    # If -I is not passed from cmdline and config file also do not have index directory path, create intree index.
    if args.index_path == "" and config_idxPath == "":
        if (args.debug_mode == "1"):
            print(("Update the intree index: ", args.fs_path))
        exec_path = "%s/bin/index/bee-update" % config.executable()
        query_cmd = [exec_path, args.fs_path, args.fs_path, mnt_path, port, str(args.debug_mode)]

    else:
        # If -I is not passed but config file has index directory path, use that path for index directory.
        if args.index_path == "":
            args.index_path = config_idxPath
        if (args.debug_mode == "1"):
            print(("Update the outtree index: ", args.index_path))
        exec_path = "%s/bin/index/bee-update" % config.executable()
        query_cmd = [exec_path, args.fs_path, args.index_path, mnt_path, port, str(args.debug_mode)]
        print(query_cmd)

    if args.no_beegfs_meta:
        query_cmd += ['-B']

    query = subprocess.Popen(query_cmd)

    query.communicate() # block until query finishes

    return query.returncode
