#!/usr/bin/env python3
# This file is part of GUFI, which is part of MarFS, which is released
# under the BSD license.
#
#
# Copyright (c) 2017, Los Alamos National Security (LANS), LLC
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation and/or
# other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#
# From Los Alamos National Security, LLC:
# LA-CC-15-039
#
# Copyright (c) 2017, Los Alamos National Security, LLC All rights reserved.
# Copyright 2017. Los Alamos National Security, LLC. This software was produced
# under U.S. Government contract DE-AC52-06NA25396 for Los Alamos National
# Laboratory (LANL), which is operated by Los Alamos National Security, LLC for
# the U.S. Department of Energy. The U.S. Government has rights to use,
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR LOS
# ALAMOS NATIONAL SECURITY, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is
# modified to produce derivative works, such modified software should be
# clearly marked, so as not to confuse it with the version available from
# LANL.
#
# THIS SOFTWARE IS PROVIDED BY LOS ALAMOS NATIONAL SECURITY, LLC AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL LOS ALAMOS NATIONAL SECURITY, LLC OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
# OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
# OF SUCH DAMAGE.



import argparse
import itertools
import os
import subprocess
import sys

import bee_common
import bee_config

# hiding traceback
sys.tracebacklimit=0

# Examples are outputs generated by running stats
# on the index of the tree generated by test/generatetree

# allowed keywords for sorting output (if applicable)
ASCENDING  = 'ASC'
DESCENDING = 'DESC'
ORDER = {'ASC'   : ASCENDING,
         'DESC'  : DESCENDING,
         'least' : ASCENDING,
         'most'  : DESCENDING}

def build_create(name, columns):
    return 'CREATE TABLE {0}({1})'.format(name, ', '.join(columns))

def build_where(args, root_uid = 0, root_gid = 0):
    where = []

    # if not ((os.geteuid() == root_uid) or (os.getegid() == root_gid)):
    #     where  += ['(uid == {})'.format(os.getuid())]
    #     # where  += ['(uid == {}) OR (gid in {})'.format(os.getuid(), '(' + str(os.getgroups())[1:-1] + ')')]

    if args.uid:
        where += ['uid == "{}"'.format(args.uid)]

    return where

def depth(config, args, where):
    '''
    get the depths of the provided directories relative to the root directory

    using --recursive/-r here will search for all directories under the provided directory as well

    $ bee stats    depth
    0

    $ bee stats -r depth
    . 0
    directory 1
    directory/subdirectory 2
    leaf_directory 1
    '''

    if(cfg_path == bee_config.DEFAULT_PATH or cfg_path == test_config_path):
        root_slash_count = idxPath.count('/')
        root_len = len(idxPath)
    else:
        root_slash_count = cfg_path.count('/')
        root_len = len(cfg_path)

    queries = [
        '-I', build_create(args.inmemory_name, ['name TEXT', 'level INT64']),

        '-S', 'INSERT INTO {} {} '.format(args.inmemory_name,
                                          bee_common.build_query(['path()', '0'],
                                                                  ['summary'],
                                                                  where,
                                                                  None,
                                                                  None,
                                                                  args.num_results,
                                                                  None)),

        '-J', 'INSERT INTO aggregate.{} {} '.format(args.inmemory_name,
                                                    bee_common.build_query(['substr(name, {})'.format(root_len + 2),
                                                                             '(length(name) - length(replace(name, "/", ""))) - {}'.format(root_slash_count)],
                                                                            [args.inmemory_name],
                                                                            None,
                                                                            None,
                                                                            None,
                                                                            None,
                                                                            None)),

        '-G', bee_common.build_query((['CASE length(name) WHEN 0 THEN "." ELSE name END'] if args.recursive else []) +
                                      ['level'],
                                      [args.inmemory_name],
                                      None,
                                      None,
                                      ['name ' + ORDER[args.order]] if args.order else None,
                                      args.num_results,
                                      None),
    ]

    if not args.recursive:
        queries += ['-y', '0',
                    '-z', '0']

    return queries

def filesize(config, args, where):
    '''
    get the size the files in the immediate directory

    using --recursive here will get the size taken up by all of files in all subdirectories, without adding them

    $ bee stats    filesize
    1049600

    $ bee stats -r filesize
    . 1049600
    directory 0
    directory/subdirectory 0
    leaf_directory 0
    '''

    if(cfg_path == bee_config.DEFAULT_PATH or cfg_path == test_config_path):
        root_len = len(idxPath)
    else:
        root_len = len(cfg_path)


    queries = [
        '-I', build_create(args.inmemory_name, ['name TEXT', 'size INT64']),

        '-E', 'INSERT INTO {} {}'.format(args.inmemory_name,
                                          bee_common.build_query(['path()', 'CAST(TOTAL(size) as INT64)'],
                                                                  ['entries'],
                                                                  where + ['type == "f"'],
                                                                  None,
                                                                  None,
                                                                  args.num_results,
                                                                  None)),

        '-J', 'INSERT INTO aggregate.{} {}'.format(args.inmemory_name,
                                                   bee_common.build_query(['substr(name, {})'.format(root_len + 2), 'size'],
                                                                           [args.inmemory_name],
                                                                           None,
                                                                           None,
                                                                           ['name ' + ORDER[args.order]] if args.order else None,
                                                                           args.num_results,
                                                                           None)),

        '-G', bee_common.build_query((['CASE length(name) WHEN 0 THEN "." ELSE name END'] if args.recursive else []) +
                                      ['size'],
                                      [args.inmemory_name],
                                      None,
                                      None,
                                      ['name ' + ORDER[args.order]] if args.order else None,
                                      args.num_results,
                                      None),
    ]

    if not args.recursive:
        queries += ['-y', '0',
                    '-z', '0']

    return queries

def entries_count(config, args, where, type):
    '''
    get the number of <type>s in the immediate directory

    using --recursive here will get the number of <type> in all subdirectories, without adding them

    $ bee stats    filecount
    7

    $ bee stats -r filecount
    . 7
    directory 3
    directory/subdirectory 1
    leaf_directory 2

    $ bee stats    linkcount
    1

    $ bee stats -r linkcount
    . 1
    directory 0
    directory/subdirectory 1
    leaf_directory 0
    '''

    if(cfg_path == bee_config.DEFAULT_PATH or cfg_path == test_config_path):
        root_len = len(idxPath)
    else:
        root_len = len(cfg_path)

    queries = [
        '-I', build_create(args.inmemory_name, ['name TEXT', 'count INT64']),

        '-E', 'INSERT INTO {} {}'.format(args.inmemory_name,
                                         bee_common.build_query(['path()', 'COUNT(*)'],
                                                                 ['entries'],
                                                                 where + ['type == "{}"'.format(type)],
                                                                 None,
                                                                 None,
                                                                 args.num_results,
                                                                 None)),

        '-J', 'INSERT INTO aggregate.{} {}'.format(args.inmemory_name,
                                                   bee_common.build_query(['substr(name, {})'.format(root_len + 2), 'count'],
                                                                           [args.inmemory_name],
                                                                           None,
                                                                           None,
                                                                           ['name ' + ORDER[args.order]] if args.order else None,
                                                                           args.num_results,
                                                                           None)),

        '-G', bee_common.build_query((['CASE length(name) WHEN 0 THEN "." ELSE name END'] if args.recursive else []) +
                                      ['count'],
                                      [args.inmemory_name],
                                      None,
                                      None,
                                      ['name ' + ORDER[args.order]] if args.order else None,
                                      args.num_results,
                                      None),
    ]

    if not args.recursive:
        queries += ['-y', '0',
                    '-z', '0']

    return queries

def filecount(config, args, where):
    return entries_count(config, args, where, 'f')

def linkcount(config, args, where):
    return entries_count(config, args, where, 'l')

def total_filesize(config, args, where):
    '''
    get the total size taken up by the entire directory

    --cumulative collects all values and sums them instead of reporting by uid

    $ bee stats    total-filesize
    1001 1049600

    $ bee stats -c total-filesize
    1049600
    '''

    if(cfg_path == bee_config.DEFAULT_PATH or cfg_path == test_config_path):
        root_len = len(idxPath)
    else:
        root_len = len(cfg_path)

    queries = []
    if args.cumulative:
        queries = [
            '-I',
            build_create(args.inmemory_name,
                         ['id INTEGER PRIMARY KEY',
                          'size INT64']),

            '-E',
            'INSERT INTO {} {}'.format(args.inmemory_name,
                                       bee_common.build_query(['NULL', 'totsize'],
                                                               ['summary'],
                                                               where,
                                                               None,
                                                               None,
                                                               None,
                                                               None)),

            '-J',
            'INSERT INTO aggregate.{} {}'.format(args.inmemory_name,
                                                 bee_common.build_query(['NULL', 'size'],
                                                                         [args.inmemory_name],
                                                                         None,
                                                                         None,
                                                                         None,
                                                                         None,
                                                                         None)),

            '-G',
            bee_common.build_query(['SUM(size)'],
                                    [args.inmemory_name],
                                    None,
                                    None,
                                    None,
                                    None,
                                    None),
        ]

    else:
        group_by = ['uid']
        order_by = ['uid {}' .format(ORDER[args.order]),
                    'size {}'.format(ORDER[args.order])]

        queries = [
            '-I',
            build_create(args.inmemory_name,
                         ['id INTEGER PRIMARY KEY',
                          'uid INT64',
                          'size INT64']),

            '-E',
            'INSERT INTO {} {}'.format(args.inmemory_name,
                                       bee_common.build_query(['NULL', 'uid', 'SUM(size)'],
                                                               ['pentries'],
                                                               where + ['type == "f"'],
                                                               group_by,
                                                               order_by,
                                                               args.num_results,
                                                               None)),

            '-J',
            'INSERT INTO aggregate.{} {}'.format(args.inmemory_name,
                                                 bee_common.build_query(['NULL', 'uid', 'size'],
                                                                         [args.inmemory_name],
                                                                         None,
                                                                         None,
                                                                         order_by,
                                                                         args.num_results,
                                                                         None)),

            '-G',
            bee_common.build_query(['uidtouser(uid, 0)', 'SUM(size)'],
                                    [args.inmemory_name],
                                    None,
                                    group_by,
                                    order_by,
                                    args.num_results,
                                    None),
        ]

    return queries

def total_entries_count(config, args, where, type):
    '''
    get the total number of <type>s under the directory

    --cumulative collects all values and sums them instead of reporting by uid

    $ bee stats    total-filecount
    1001 13

    $ bee stats -c total-filecount
    13

    $ bee stats    total-linkcount
    1001 2

    $ bee stats -c total-linkcount
    2
    '''

    group_by = None if args.cumulative else ['uid']

    queries = [
        '-I', build_create(args.inmemory_name, ['id INTEGER PRIMARY KEY', 'uid INT64', 'count INT64']),

        '-E', 'INSERT INTO {} {}'.format(args.inmemory_name,
                                         bee_common.build_query(['NULL', 'uid', 'COUNT(*) AS count'],
                                                                 ['entries'],
                                                                 where + ['type == "{}"'.format(type)],
                                                                 group_by,
                                                                 ['count {}'.format(ORDER[args.order]), 'uid {}'.format(ORDER[args.order])],
                                                                 args.num_results,
                                                                 None)),

        '-J', 'INSERT INTO aggregate.{} {}'.format(args.inmemory_name,
                                                   bee_common.build_query(['NULL', 'uid', 'count'],
                                                                           [args.inmemory_name],
                                                                           None,
                                                                           None,
                                                                           ['count {}'.format(ORDER[args.order]), 'uid {}'.format(ORDER[args.order])],
                                                                           args.num_results,
                                                                           None)),

        '-G', bee_common.build_query(([] if args.cumulative else ['uidtouser(uid, 0)']) + ['SUM(count) AS count'],
                                      [args.inmemory_name],
                                      None,
                                      group_by,
                                      ['count {}'.format(ORDER[args.order]), 'uid {}'.format(ORDER[args.order])],
                                      args.num_results,
                                      None),
    ]

    return queries

def total_filecount(config, args, where):
    return total_entries_count(config, args, where, 'f')

def total_linkcount(config, args, where):
    return total_entries_count(config, args, where, 'l')

def dircount(config, args, where):
    '''
    get the number of directories in the immediate directory

    using --recursive here will get the number of directoriess in all subdirectories, without adding them

    ############################################
    # This is wrong when running as a non-root #
    # user. Without root, the other databases  #
    # can't be accessed.                       #
    ############################################

    $ bee stats    dircount
    2

    $ bee stats -r dircount
    . 2
    directory 1
    directory/subdirectory 0
    leaf_directory 0
    '''

    if(cfg_path == bee_config.DEFAULT_PATH or cfg_path == test_config_path):
        root_len = len(idxPath)
    else:
        root_len = len(cfg_path)


    queries = [
        '-I', build_create(args.inmemory_name, ['path TEXT', 'inode TEXT', 'pinode TEXT']),

        '-S', 'INSERT INTO {} {}'.format(args.inmemory_name,
                                         bee_common.build_query(['fpath()', 'inode', 'pinode'],
                                                                 ['summary'],
                                                                 where,
                                                                 None,
                                                                 None,
                                                                 None)),

        '-J', 'INSERT INTO aggregate.{0} SELECT * FROM {0}'.format(args.inmemory_name),

        '-G', bee_common.build_query((['CASE (length(parent.path) - {0}) WHEN 0 THEN "." ELSE substr(parent.path, {1}) END'.format(root_len, root_len + 2)] if args.recursive else []) +
                                      ['IFNULL(COUNT(child.pinode), 0) END'],
                                      ['out parent LEFT JOIN out child On parent.inode == child.pinode'],
                                      where + ([] if args.recursive else ['parent.path == "{}"'.format(args.path)]),
                                      ['parent.inode'],
                                      ['parent.path'],
                                      args.num_results,
                                      None),
    ]

    if not args.recursive:
        queries += [
            '-y', '0',
            '-z', '1']

    return queries

def total_dircount(config, args, where):
    '''
    get the total number of directories under the directory

    --cumulative collects all values and sums them instead of reporting by uid

    $ bee stats    total-dircount
    1001 3

    $ bee stats -c total-dircount
    3
    '''

    group_by = None if args.cumulative else ['uid']

    queries = [
        '-I', build_create(args.inmemory_name, ['id INTEGER PRIMARY KEY', 'uid INT64', 'count INT64']),

        '-S', 'INSERT INTO {} {}'.format(args.inmemory_name,
                                         bee_common.build_query(['NULL', 'uid', '1 as count'],
                                                                 ['summary'],
                                                                 where + ['level() > 0'],
                                                                 group_by,
                                                                 ['count {}'.format(ORDER[args.order]), 'uid {}'.format(ORDER[args.order])],
                                                                 args.num_results,
                                                                 None)),

        '-J', 'INSERT INTO aggregate.{} {}'.format(args.inmemory_name,
                                                   bee_common.build_query(['NULL', 'uid', 'count'],
                                                                           [args.inmemory_name],
                                                                           None,
                                                                           None,
                                                                           ['count {}'.format(ORDER[args.order]), 'uid {}'.format(ORDER[args.order])],
                                                                           args.num_results,
                                                                           None)),

        '-G', bee_common.build_query(([] if args.cumulative else ['uidtouser(uid, 0)']) + ['SUM(count) AS count'],
                                      [args.inmemory_name],
                                      None,
                                      group_by,
                                      ['count {}'.format(ORDER[args.order]), 'uid {}'.format(ORDER[args.order])],
                                      args.num_results,
                                      None),
    ]

    return queries

def leaf_dirs(config, args, where):
    '''
    get the leaf directories immediately under the current directory

    using --recursive/-r here will search for leaf directories under the provided directory as well

    $ bee stats    leaf-dirs
    leaf_directory

    $ bee stats -r leaf-dirs
    directory/subdirectory
    leaf_directory
    '''

    if(cfg_path == bee_config.DEFAULT_PATH or cfg_path == test_config_path):
        root_len = len(idxPath)
    else:
        root_len = len(cfg_path)

    queries = [
        '-I', build_create(args.inmemory_name, ['name TEXT']),

        '-S', 'INSERT INTO {} {}'.format(args.inmemory_name,
                                         bee_common.build_query(['path()'],
                                                                 ['summary'],
                                                                 where + ['nlink == 2'],
                                                                 None,
                                                                 None,
                                                                 args.num_results,
                                                                 None)),

        '-J', 'INSERT INTO aggregate.{} {}'.format(args.inmemory_name,
                                                   bee_common.build_query(['substr(name, {})'.format(root_len + 2)],
                                                                           [args.inmemory_name],
                                                                           None,
                                                                           None,
                                                                           None,
                                                                           None,
                                                                           None)),

        '-G', bee_common.build_query(['CASE length(name) WHEN 0 THEN "." ELSE name END'],
                                      [args.inmemory_name],
                                      None,
                                      None,
                                      ['name ' + ORDER[args.order]] if args.order else None,
                                      args.num_results,
                                      None),
    ]

    if not args.recursive:
        queries += ['-y', '1',
                    '-z', '1']

    return queries

def leaf_depth(config, args, where):
    '''
    get the leaf directories immediately under the current directory

    using --recursive/-r here will search for leaf directories under the provided directory as well

    $ bee stats    leaf-depth
    leaf_directory 1

    $ bee stats -r leaf-depth
    directory/subdirectory 2
    leaf_directory 1
    '''

    if(cfg_path == bee_config.DEFAULT_PATH or cfg_path == test_config_path):
        root_slash_count = idxPath.count('/')
        root_len = len(idxPath)

    else:
        root_slash_count = cfg_path.count('/')
        root_len = len(cfg_path)

    queries = [
        '-I', build_create(args.inmemory_name, ['name TEXT', 'level INT64']),

        '-S', 'INSERT INTO {} {}'.format(args.inmemory_name,
                                         bee_common.build_query(['path()', '0'],
                                                                 ['summary'],
                                                                 where + ['nlink == 2'],
                                                                 None,
                                                                 None,
                                                                 args.num_results,
                                                                 None)),

        '-J', 'INSERT INTO aggregate.{} {}'.format(args.inmemory_name,
                                                   bee_common.build_query(['substr(name, {})'.format(root_len + 2),
                                                                            '(length(name) - length(replace(name, "/", ""))) - {}'.format(root_slash_count)],
                                                                           [args.inmemory_name],
                                                                           None,
                                                                           None,
                                                                           None,
                                                                           None,
                                                                           None)),

        '-G', bee_common.build_query(['name', 'level'],
                                      [args.inmemory_name],
                                      None,
                                      None,
                                      ['name ' + ORDER[args.order]] if args.order else None,
                                      args.num_results,
                                      None),
    ]

    if not args.recursive:
        queries += ['-y', '1',
                    '-z', '1']

    return queries

def leaf(config, args, where, type):
    '''
    get number of <type>s in the leaf directories immediately under the current directory

    using --recursive/-r here will search for leaf directories under the provided directory as well

    $ bee stats    leaf-files
    leaf_directory 2

    $ bee stats -r leaf-files
    directory/subdirectory 1
    leaf_directory 2
    '''

    if(cfg_path == bee_config.DEFAULT_PATH or cfg_path == test_config_path):
        root_len = len(idxPath)
    else:
        root_len = len(cfg_path)

    queries = [
        '-I', build_create(args.inmemory_name, ['name TEXT', 'count INT64']),

        '-S', 'INSERT INTO {} {}'.format(args.inmemory_name,
                                         bee_common.build_query(['path()', '(SELECT COUNT(*) FROM entries WHERE type=="{}")'.format(type)],
                                                                 ['summary'],
                                                                 where + ['nlink == 2'],
                                                                 None,
                                                                 None,
                                                                 args.num_results,
                                                                 None)),

        '-J', 'INSERT INTO aggregate.out ' + bee_common.build_query(['substr(name, {})'.format(root_len + 2),
                                                                      'count'],
                                                                     [args.inmemory_name],
                                                                     None,
                                                                     None,
                                                                     None,
                                                                     None,
                                                                     None),

        '-G',                                bee_common.build_query(['name', 'count'],
                                                                     [args.inmemory_name],
                                                                     None,
                                                                     None,
                                                                     ['name ' + ORDER[args.order]] if args.order else None,
                                                                     args.num_results,
                                                                     None),
    ]

    if not args.recursive:
        queries += ['-y', '1',
                    '-z', '1']

    return queries

def leaf_files(config, args, where):
    return leaf(config, args, where, 'f')

def leaf_links(config, args, where):
    return leaf(config, args, where, 'l')

def total_leaf(config, args, where, type):
    '''
    get total number of <type>s in the leaf directories under this directory

    --cumulative collects all values and sums them instead of reporting by uid

    $ bee stats    total-leaf-files
    1001 3

    $ bee stats -c total-leaf-files
    3
    '''

    group_by = None if args.cumulative else ['uid']

    queries = [
        '-I', build_create(args.inmemory_name, ['uid INT64', 'count INT64']),

        '-S', 'INSERT INTO {} {}'.format(args.inmemory_name,
                                         bee_common.build_query(['uid', '(SELECT COUNT(*) FROM entries WHERE type=="{}") AS count'.format(type)],
                                                                 ['summary'],
                                                                 where + ['nlink == 2'],
                                                                 group_by,
                                                                 ['uid {}'.format(ORDER[args.order]), 'count {}'.format(ORDER[args.order])],
                                                                 args.num_results,
                                                                 None)),

        '-J', 'INSERT INTO aggregate.{} {}'.format(args.inmemory_name,
                                                   bee_common.build_query(['uid', 'SUM(count) AS count'],
                                                                           [args.inmemory_name],
                                                                           None,
                                                                           group_by,
                                                                           ['uid {}'.format(ORDER[args.order]), 'count {}'.format(ORDER[args.order])],
                                                                           args.num_results,
                                                                           None)),

        '-G', bee_common.build_query(([] if args.cumulative else ['uidtouser(uid, 0)']) + ['SUM(count) AS count'],
                                      [args.inmemory_name],
                                      None,
                                      group_by,
                                      ['uid {}'.format(ORDER[args.order]), 'count {}'.format(ORDER[args.order])],
                                      args.num_results,
                                      None),
    ]

    return queries

def total_leaf_files(config, args, where):
    return total_leaf(config, args, where, 'f');

def total_leaf_links(config, args, where):
    return total_leaf(config, args, where, 'l');

def median_leaf_files(config, args, where):
    '''
    get median number of leaf files under the provided directory

    $ bee stats median-leaf-files
    1.5
    '''

    queries = [
        '-I', build_create(args.inmemory_name, ['uid INT64', 'count INT64']),

        '-S', 'INSERT INTO {} {}'.format(args.inmemory_name,
                                         bee_common.build_query(['uid', '(SELECT COUNT(*) FROM entries WHERE type=="f") AS count'],
                                                                 ['summary'],
                                                                 where + ['nlink == 2'],
                                                                 None,
                                                                 None,
                                                                 args.num_results,
                                                                 None)),

        '-J', 'INSERT INTO aggregate.{} {}'.format(args.inmemory_name,
                                                   bee_common.build_query(['uid', 'count'],
                                                                           [args.inmemory_name],
                                                                           None,
                                                                           None,
                                                                           None,
                                                                           None)),

        # Adapter from answer by CL
        # https://stackoverflow.com/a/15766121
        '-G', '''SELECT AVG(count)
                 FROM (SELECT count
                       FROM {0}
                       ORDER BY count
                       LIMIT 2 - (SELECT COUNT(*) FROM {0}) % 2    -- odd 1, even 2
                                  OFFSET (SELECT (COUNT(*) - 1) / 2
                                          FROM {0})
                      )'''.format(args.inmemory_name)
    ]

    return queries

def average_leaf(config, args, where, type):
    '''
    get average number of leaf files under the provided directory
    '''

    group_by = ['uid'] if args.cumulative else None

    count = bee_common.build_query(['COUNT(*)'],
                                    ['entries'],
                                    ['type == "{}"'.format(type)],
                                    group_by,
                                    None,
                                    None)

    queries = [
        '-I', build_create(args.inmemory_name, ['id INTEGER PRIMARY KEY', 'uid INT64', 'count INT64']),

        '-S', 'INSERT INTO {} {}'.format(args.inmemory_name,
                                         bee_common.build_query(['NULL',
                                                                  'uid',
                                                                  '({}) AS count'.format(count)
                                                              ],
                                                                 ['summary'],
                                                                 where + ['nlink == 2'],
                                                                 None,
                                                                 None,
                                                                 args.num_results,
                                                                 None)),

        '-J', 'INSERT INTO aggregate.{} {}'.format(args.inmemory_name,
                                                   bee_common.build_query(['NULL', 'uid', 'count'],
                                                                           [args.inmemory_name],
                                                                           None,
                                                                           None,
                                                                           None,
                                                                           None)),

        # Adapter from answer by CL
        # https://stackoverflow.com/a/15766121
        '-G', bee_common.build_query(['AVG(count)'],
                                      [args.inmemory_name],
                                      None,
                                      group_by,
                                      None,
                                      args.num_results,
                                      None)
    ]

    return queries

def average_leaf_files(config, args, where):
    return average_leaf(config, args, where, 'f')

def average_leaf_links(config, args, where):
    return average_leaf(config, args, where, 'l')

def per_level(config, args, where, type):
    '''
    get counts of how many <type>s are in each level of the tree, by uid

    --cumulative combines the results into one set of distributions

    $ bee stats    files-per-level
    1001 0 7
    1001 1 5
    1001 2 1

    $ bee stats -c files-per-level
    0 7
    1 5
    2 1

    $ bee stats    links-per-level
    1001 0 1
    1001 2 1

    $ bee stats -c links-per-level
    0 1
    2 1

    $ bee stats    dirs-per-level
    1001 0 1
    1001 1 2
    1001 2 1

    $ bee stats -c dirs-per-level
    0 1
    1 2
    2 1
    '''

    if(cfg_path == bee_config.DEFAULT_PATH or cfg_path == test_config_path):
        root = os.path.normpath(idxPath)
    else:
        root = os.path.normpath(cfg_path)

    root_len = len(root)
    path = args.path
    path_len = len(path)
    starting_depth = 0
    if path_len != root_len:
        starting_depth = path[root_len - path_len:].count('/')

    group_by = ['level'] + ([] if args.cumulative else ['uid'])
    order_by = None if args.cumulative else ['level {}'.format(ORDER[args.order]), 'uid {}'.format(ORDER[args.order])]

    uid = [] if args.cumulative else ['uid']
    count = '1' if type == 'd' else '(SELECT COUNT(*) FROM entries WHERE type == "{}")'.format(type)

    queries = [
        '-I', build_create(args.inmemory_name, ['id INT PRIMARY KEY'] + ([] if args.cumulative else ['uid INT64']) + ['level INT64', 'count INT64']),

        '-S', 'INSERT INTO {} {}'.format(args.inmemory_name,
                                         bee_common.build_query(['NULL'] + uid + ['level() + {}'.format(starting_depth), count],
                                                                 ['summary'],
                                                                 where,
                                                                 ['uid'],
                                                                 None,
                                                                 None,
                                                                 None)),

        '-J', 'INSERT INTO aggregate.{} {}'.format(args.inmemory_name,
                                                   bee_common.build_query(['NULL'] + uid + ['level', 'SUM(count)'],
                                                                           [args.inmemory_name],
                                                                           None,
                                                                           group_by,
                                                                           order_by,
                                                                           None,
                                                                           None)),

        '-G', bee_common.build_query(([] if args.cumulative else ['uidtouser(uid, 0)']) + ['level', 'SUM(count)'],
                                      [args.inmemory_name],
                                      ['count <> 0'],
                                      group_by,
                                      order_by,
                                      args.num_results,
                                      None),

    ]

    return queries

def files_per_level(config, args, where):
    return per_level(config, args, where, 'f')

def links_per_level(config, args, where):
    return per_level(config, args, where, 'l')

def dirs_per_level(config, args, where):
    return per_level(config, args, where, 'd')

def duplicate_names(config, args, where):
    columns = ['name', 'type', 'inode', 'mode', 'nlink', 'uid', 'gid', 'size', 'blksize', 'blocks', 'atime', 'mtime', 'ctime', 'linkname', 'xattrs']

    if(cfg_path == bee_config.DEFAULT_PATH or cfg_path == test_config_path):
        root = os.path.normpath(idxPath)
    else:
        root = os.path.normpath(cfg_path)

    root_len = len(root)

    queries = [
        '-I',
        build_create(args.inmemory_name, ['id INTEGER PRIMARY KEY', 'path TEXT', 'name TEXT', 'type TEXT', 'inode TEXT', 'mode INT64', 'nlink INT64', 'uid INT64', 'gid INT64', 'size INT64', 'blksize INT64', 'blocks INT64', 'atime INT64', 'mtime INT64', 'ctime INT64', 'linkname TEXT', 'xattrs TEXT']),

        '-S',
        'INSERT INTO {} {}'.format(args.inmemory_name,
                                   bee_common.build_query(['NULL', 'path()'] +
                                                           columns,
                                                           ['summary'],
                                                           where,
                                                           None,
                                                           None,
                                                           None,
                                                           None)),

        '-E',
        'INSERT INTO {} {}'.format(args.inmemory_name,
                                   bee_common.build_query(['NULL', 'path()'] +
                                                           columns,
                                                           ['entries'],
                                                           where,
                                                           None,
                                                           None,
                                                           None,
                                                           None)),

        '-J',
        'INSERT INTO aggregate.out ' + bee_common.build_query(['NULL', 'path'] +
                                                               columns,
                                                               [args.inmemory_name],
                                                               None,
                                                               None,
                                                               None,
                                                               None,
                                                               None),

        '-G',
        'SELECT lhs.name, lhs.path FROM {} as lhs INNER JOIN (SELECT DISTINCT name, type, size FROM {} GROUP BY name, type, size HAVING COUNT(*) > 1) as rhs ON ((lhs.name == rhs.name) AND (lhs.type == rhs.type) AND (lhs.size == rhs.size)) ORDER BY lhs.name'.format(args.inmemory_name, args.inmemory_name) + ('LIMIT {}'.format if args.num_results else '')

    ]

    queries += ['-a']

    return queries

# argv[0] should be the command name
def exec_stats(argv, config_path):
    # mapping of desired statistics to functions
    # uses --recursive
    RECURSIVE = {
        'depth'              : depth,
        'filesize'           : filesize,
        'filecount'          : filecount,
        'linkcount'          : linkcount,
        'dircount'           : dircount,
        'leaf-dirs'          : leaf_dirs,
        'leaf-depth'         : leaf_depth,
        'leaf-files'         : leaf_files,
        'leaf-links'         : leaf_links,
    }

    # uses --cumulative
    CUMULATIVE = {
        'total-filesize'     : total_filesize,
        'total-filecount'    : total_filecount,
        'total-linkcount'    : total_linkcount,
        'total-dircount'     : total_dircount,
        'total-leaf-files'   : total_leaf_files,
        'total-leaf-links'   : total_leaf_links,
        'files-per-level'    : files_per_level,
        'links-per-level'    : links_per_level,
        'dirs-per-level'     : dirs_per_level,
        'average-leaf-files' : average_leaf_files,
        'average-leaf-links' : average_leaf_links,
    }

    # does not use either
    OTHERS = {
        'median-leaf-files'  : median_leaf_files,
        'duplicate-names'    : duplicate_names,
    }

    STATS = {}
    STATS.update(RECURSIVE)
    STATS.update(CUMULATIVE)
    STATS.update(OTHERS)

    global cfg_path
    cfg_path = config_path

    # storing name of test config file
    global test_config_path
    test_config_path = "config.test"

    # find and parse the configuration file first
    if(config_path == bee_config.DEFAULT_PATH or config_path == test_config_path):
        config = bee_config.Server(config_path)
    else:
        config = bee_config.envServer(bee_config.DEFAULT_PATH)

    # parse the arguments
    parser = argparse.ArgumentParser(description='BeeGFS Hive index statistics', add_help=False)

    # override help to not use -h
    parser.add_argument('--help',                                                   action='help',                                         help='show this help message and exit')
    parser.add_argument('--version', '-v',                                          action='version',          version=os.path.basename(os.path.realpath(__file__)) + ' 2.0.0')
    parser.add_argument('--recursive', '-r',                  dest='recursive',     action='store_true',                                   help='run command recursively ({})'.format(', '.join(list(RECURSIVE.keys()))))
    parser.add_argument('--cumulative', '-c',                 dest='cumulative',    action='store_true',                                   help='return cumulative values ({})'.format(', '.join(list(CUMULATIVE.keys()))))
    parser.add_argument('--order',           metavar='order', dest='order',         choices=list(ORDER.keys()),               default=ASCENDING, help='sort output (if applicable)')
    parser.add_argument('--delim',           metavar='c',     dest='delim',         type=bee_common.get_char,          default=' ',       help='delimiter separating output columns')
    parser.add_argument('--num-results',     metavar='n',     dest='num_results',   type=bee_common.get_non_negative,                     help='first n results')
    parser.add_argument('--uid', '--user',   metavar='u',     dest='uid',           type=bee_common.get_uid,                              help='restrict to user')
    parser.add_argument('--in-memory-name',  metavar='name',  dest='inmemory_name', type=str,                           default='out',    help=argparse.SUPPRESS)
    parser.add_argument('stat',                                                     choices=list(STATS.keys()), help='statistic to get')
    parser.add_argument('path',                                                     type=str, nargs='?',                default='')

    args = parser.parse_args(argv[2:]);

    srcpath = []
    indexpaths = []
    global idxPath

    # check args
    if args.recursive and (args.stat in CUMULATIVE):
        sys.stderr.write('--recursive/-r has no effect on "{}" statistic\n'.format(args.stat))
    if args.cumulative and (args.stat in RECURSIVE):
        sys.stderr.write('--cumulative/-c has no effect on "{}" statistic\n'.format(args.stat))
    if (args.recursive or args.cumulative) and (args.stat in OTHERS):
        sys.stderr.write('--recursive/-r and --cumulative/-c have no effect on "{}" statistic\n'.format(args.stat))

    # prepend the provided paths int the config file as a filesystem path
    config_paths = config.indexpaths().split(':')
    srcpath = config_paths[0]
    if len(config_paths) == 2:
        idxPath = config_paths[1]
    else:
        idxPath = config_paths[0]

    if len(args.path) == 0 or args.path[0] != "/":
        current_dir = os.getcwd()
        path1 = "%s/%s" % (current_dir, args.path)
    else:
        path1 = args.path

    # check is path is part of index or FSpath
    if not path1.startswith(idxPath) and not path1.startswith(srcpath):
        print("Error: Invalid path. Please run from index or FSPath directory")
        exit(1)

    # If the path starts with fspath, replace it with index
    if path1.startswith(srcpath):
        path1 = path1.replace(srcpath, idxPath)

    # create the query command
    exec_path = "%s/bin/index/bee-search" % config.executable()
    query_cmd = [exec_path,
                 '-e', '0',
                 '-n', str(config.threads()),
                 '-d', args.delim] + STATS[args.stat](config, args, build_where(args))

    query = subprocess.Popen(query_cmd + [path1])
    query.communicate() # block until query finishes

    return query.returncode
